from typing import Any, Dict, List, Type  # noqa: F401

import shotgun_api3
import shotgun_api3.lib.mockgun

#: Entity plugins that are registered to pyshotgrid.
__ENTITY_PLUGINS = []  # type: List[Dict[str,Any]]
#: Fallback entity class that is used when no match in the __ENTITY_PLUGINS is found.
__ENTITY_FALLBACK_CLASS = None  # type: Type|None
#: The class that represents the ShotGrid site.
__SG_SITE_CLASS = None  # type: Type|None


def new_entity(sg, *args, **kwargs):
    """
    Create a new instance of a pyshotgrid class that represents a ShotGrid entity.
    This function is meant to be used as the main way to create new pyshotgrid instances
    and will always return the correct entity instance that you should work with.

    .. Note::

        This does NOT *create* an entity in Shotgrid. It just gives you a python
        object that *represents* an entity.

    The function can be used in 3 ways which all do the same thing::

        sg_entity = pyshotgrid.new_entity(sg, {'type': 'Project', 'id': 1})
        sg_entity = pyshotgrid.new_entity(sg, 'Project', 1)
        sg_entity = pyshotgrid.new_entity(sg, entity_type='Project', entity_id=1)

    :param shotgun_api3.shotgun.Shotgun sg: A fully initialized Shotgun instance.
    :return: The pyshotgrid object or None if it could not be converted.
    :rtype: SGEntity|None
    """
    entity_type = None
    entity_id = None
    if args:
        if isinstance(args[0], dict):
            # new_entity(sg, {'type': 'Project', 'id': 1})
            entity_type = args[0]["type"]
            entity_id = args[0]["id"]
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], int):
            # new_entity(sg, 'Project', 1)
            entity_type = args[0]
            entity_id = args[1]
    elif kwargs:
        if "entity_type" in kwargs and "entity_id" in kwargs:
            entity_type = kwargs["entity_type"]
            entity_id = kwargs["entity_id"]

    if entity_type is not None and entity_id is not None:

        for entity_plugin in __ENTITY_PLUGINS:
            if entity_type in [
                entity_plugin["shotgrid_type"],
                entity_plugin["display_name"],
            ]:
                return entity_plugin["pysg_class"](sg, entity_id)
        # noinspection PyCallingNonCallable
        return __ENTITY_FALLBACK_CLASS(sg, entity_type, entity_id)


def new_site(*args, **kwargs):
    """
    This function will create a new :py:class:`pyshotgrid.SGSite <pyshotgrid.sg_site.SGSite>`
    instance that represents a ShotGrid site.
    You can pass in either a shotgun_api3.Shotgun instance or
    the parameters of shotgun_api3.Shotgun itself. So this is equivalent::

        >>> sg = shotgun_api3.Shotgun(base_url='https://example.shotgunstudio.com',
        ...                           script_name='Some User',
        ...                           api_key='$ome_password')
        >>> sg_site = new_site(sg)

        >>> sg_site = new_site(base_url='https://example.shotgunstudio.com',
        ...                    script_name='Some User',
        ...                    api_key='$ome_password')

    :return: A new instance of the pyshotgrid site.
    :rtype: SGSite|None
    """
    if args:
        if isinstance(
            args[0], (shotgun_api3.Shotgun, shotgun_api3.lib.mockgun.Shotgun)
        ):
            sg = args[0]
        else:
            sg = shotgun_api3.Shotgun(*args)
    else:
        sg = shotgun_api3.Shotgun(**kwargs)
    return __SG_SITE_CLASS(sg)


def register_pysg_class(shotgrid_type, pysg_class, display_name=None):
    """
    Register a class for a ShotGrid type to pyshotgrid.
    This is best illustrated as by an example: Suppose you have a custom entity setup where you
    have an Episode in each project that collects some sequences of shots. It would be nice to have
    some additional functionality on the ShotGridEntity for episode objects (like a "sequences"
    function that returns all the sequences belonging to that episode). What you would do is to
    create a class SGEpisode that inherits from ShotGridEntity and add all the functionality you
    like to it. After that you call::

        register_plugin(shotgrid_type="CustomProjectEntity01",
                        pysg_class=SGEpisode,
                        display_name='Episode')

    This will register the class to pyshotgrid and the `new_entity` function will automatically
    create an "SGEpisode" instance as soon as it encounters an Episode entity. This is also
    true for all queries that happen from a ShotGridEntity. So `sg_project['sg_episodes']` would
    return "SGEpisode" instances as well.

    .. Note::

        Registering a class for an existing entity will overwrite the existing entity class.
        This way you can add/overwrite functionality for the classes that are shipped by default.

    :param str shotgrid_type: The ShotGrid entity type to register for.
    :param class pysg_class: The class to use for this entity type.
    :param str|None display_name: The display name of the entity type. If this is None, the display
                                  name will be the same as the shotgrid_type parameter.
    """
    global __ENTITY_PLUGINS

    __ENTITY_PLUGINS.append(
        {
            "shotgrid_type": shotgrid_type,
            "pysg_class": pysg_class,
            "display_name": display_name or shotgrid_type,
        }
    )


def register_fallback_pysg_class(pysg_class):
    """
    Register a class for shotgrid entities that will be used when none of the specific
    entity classes have a match.

    .. Note::

        By default this registers the SGEntity class as fallback, but you can use it to
        overwrite this behaviour.

    :param class pysg_class: The class to use as a fallback.
    """
    global __ENTITY_FALLBACK_CLASS
    __ENTITY_FALLBACK_CLASS = pysg_class


def register_sg_site_class(sg_site_class):
    """
    Register a class that represents the ShotGrid site.

    .. Note::

        This defaults to the SGSite class, but you can use it to
        overwrite this behaviour.

    :param class sg_site_class: The class to use as a fallback.
    """
    global __SG_SITE_CLASS
    __SG_SITE_CLASS = sg_site_class


def convert_fields_to_pysg(sg, fields):
    """
    Convert all the values from a fields dict to pysg objects where possible.

    :param shotgun_api3.Shotgun sg: A fully initialized Shotgun instance.
    :param dict[str,Any] fields: A fields dict as returned from a shotgun_api3.Shotgun.find()
                                 call for example.
    :return: The same dict with all values converted to pysg objects where possible.
    :rtype: dict[str,Any]
    """
    return {field: convert_value_to_pysg(sg, value) for field, value in fields.items()}


def convert_fields_to_dicts(fields):
    """
    Convert all the values from a fields dict to simple dictionaries. The counterpart function
    to `func:_convert_fields_to_pysg`.

    :param dict[str,Any] fields: A fields dict as returned from a shotgun_api3.Shotgun.find()
                                 call for example.
    :return: The same dict with all pysg objects converted to dictionaries.
    :rtype: dict[str,Any]
    """
    return {field: convert_value_to_dict(value) for field, value in fields.items()}


def convert_value_to_dict(value):
    """
    Convert any pysg objects form the given value to simple dictionaries.

    :param Any value: A field value
    :return: The value with all pysg objects converted to dictionaries.
    :rtype: dict[str,Any]
    """
    if isinstance(value, list):
        tmp = []
        for entity in value:
            if isinstance(entity, __ENTITY_FALLBACK_CLASS):
                tmp.append(entity.to_dict())
            else:
                tmp.append(entity)
        return tmp
    elif isinstance(value, __ENTITY_FALLBACK_CLASS):
        return value.to_dict()
    else:
        return value


def convert_filters_to_dict(filters):
    """
    Convert any pysg objects form the given shotgun_api3 filter to simple dictionaries.

    Example::

        >>> convert_filters_to_dict([['user', 'is', SGHumanUser(sg, id=5)]])
        [['user', 'is', {'type': 'HumanUser', 'id': 5}]]

    :param list[list] filters: The filters to convert
    :return: The filter with all pysg objects converted to dictionaries.
    :rtype: list[list]
    """
    for f in filters:
        if isinstance(f[2], list):
            tmp = []
            for entity in f[2]:
                if isinstance(entity, __ENTITY_FALLBACK_CLASS):
                    tmp.append(entity.to_dict())
                else:
                    tmp.append(entity)
            f[2] = tmp
        elif isinstance(f[2], __ENTITY_FALLBACK_CLASS):
            f[2] = f[2].to_dict()

    return filters


def convert_value_to_pysg(sg, value):
    """
    Convert the value from a field to pysg object(s) where possible.

    :param shotgun_api3.Shotgun sg: A fully initialized Shotgun instance.
    :param Any value: A field value
    :return: The value converted to pysg object(s) where possible.
    :rtype: Any
    """
    if isinstance(value, list):
        return [new_entity(sg, entity) for entity in value]
    elif isinstance(value, dict) and "type" in value and "id" in value:
        return new_entity(sg, value)
    else:
        return value



import os

from .core import convert_value_to_dict, convert_value_to_pysg, new_entity


class Field(object):
    """
    This class represents a field on a ShotGrid entity.
    It provides an interface to manage various aspects of a field.
    """

    # Note to developers:
    # The naming convention of this class intentionally leaves out the "SG" in front,
    # since there is a ShotGrid entity that is called "Field".

    def __init__(self, name, entity):
        """
        :param str name: The name of the field.
        :param SGEntity entity: The entity that this field is attached to.
        """
        self._name = name
        self._entity = entity

    @property
    def name(self):
        """
        :return: The name of the field.
        :rtype: str
        """
        return self._name

    @property
    def entity(self):
        """
        :return: The entity that this field is attached to.
        :rtype: SGEntity
        """
        return self._entity

    def get(self, raw_values=False):
        """
        :param bool raw_values: Whether to return the raw dict values or the values converted to
                                pyshotgrid objects.
        :return: The value of the field. Any entities will be automatically converted to
                 pyshotgrid objects.
        :rtype: Any
        """
        value = self._entity.sg.find_one(
            entity_type=self._entity.type,
            filters=[["id", "is", self._entity.id]],
            fields=[self._name],
        ).get(self._name)
        return value if raw_values else convert_value_to_pysg(self._entity.sg, value)

    def set(self, value):
        """
        Set the field to the given value in ShotGrid.

        :param Any value: The value to set the field to.
        """
        self._entity.sg.update(
            self._entity.type,
            self._entity.id,
            data={self._name: convert_value_to_dict(value)},
        )

    def add(self, values):
        """
        Add some values to this field

        :param list[Any] values: The value to add to this field.
        """
        self._entity.sg.update(
            self._entity.type,
            self._entity.id,
            data={self._name: convert_value_to_dict(values)},
            multi_entity_update_modes="add",
        )

    def remove(self, values):
        """
        Remove some values from this field.

        :param list[Any] values: The values to remove from this field.
        """
        self._entity.sg.update(
            self._entity.type,
            self._entity.id,
            data={self._name: convert_value_to_dict(values)},
            multi_entity_update_modes="remove",
        )

    def upload(self, path, display_name=None):
        """
        Upload a file to this field.

        :param str path: The path to the file to upload.
        :param str display_name: The display name of the file in ShotGrid.
        :return: The Attachment entity that was created for the uploaded file.
        :rtype: SGEntity
        """
        sg_attachment_id = self._entity.sg.upload(
            entity_type=self._entity.type,
            entity_id=self._entity.id,
            path=path,
            field_name=self._name,
            display_name=display_name,
        )
        return new_entity(self._entity.sg, "Attachment", sg_attachment_id)

    def download(self, path):
        """
        Download a file from a field.

        :param str path: The path to download to.
        :return:
        :raises:
            :RuntimeError: When nothing was uploaded to this field.
        """
        sg_attachment = self._entity.sg.find_one(
            self._entity.type, [["id", "is", self._entity.id]], [self._name]
        )[self._name]

        if sg_attachment is None:
            raise RuntimeError(
                'Cannot download file from field "{}" on entity "{}", because there'
                "is nothing uploaded.".format(self.name, self.entity.to_dict())
            )

        # if we can split of a file extension from the given path we assume that the path is the
        # full path with file name to download to. In the other case we assume that the path is
        # the directory to download to and attach the attachment name as the file name to the
        # directory path.
        _, ext = os.path.splitext(path)
        if ext:
            local_file_path = os.path.join(path, sg_attachment["name"])
        else:
            local_file_path = path

        return self._entity.sg.download_attachment(
            attachment=sg_attachment, file_path=local_file_path
        )

    def schema(self):
        """
        :return: The schema of this field.
                 For example::

                     {'custom_metadata': {'editable': True, 'value': ''},
                      'data_type': {'editable': False, 'value': 'status_list'},
                      'description': {'editable': True, 'value': ''},
                      'editable': {'editable': False, 'value': True},
                      'entity_type': {'editable': False, 'value': 'Shot'},
                      'mandatory': {'editable': False, 'value': False},
                      'name': {'editable': True, 'value': 'Sound delivery'},
                      'properties': {'default_value': {'editable': True,
                                                       'value': ''},
                                     'display_values': {'editable': False,
                                                        'value': {'dlvr': 'Delivered',
                                                                  'wtg': 'Waiting '
                                                                         'to '
                                                                         'Start'}},
                                     'hidden_values': {'editable': False,
                                                       'value': []},
                                     'summary_default': {'editable': True,
                                                         'value': 'none'},
                                     'valid_values': {'editable': True,
                                                      'value': ['dlvr',
                                                                'wtg']}},
                      'ui_value_displayable': {'editable': False,
                                               'value': True},
                      'unique': {'editable': False, 'value': False},
                      'visible': {'editable': True, 'value': True}}
        """
        return self._entity.sg.schema_field_read(self._entity.type, self._name)[
            self._name
        ]

    def _update_schema(self, prop, value, project_entity=None):
        """
        Update a property of the field.

        :return: True when the update succeeded.
        :rtype: bool
        """
        return self._entity.sg.schema_field_update(
            self._entity.type,
            self._name,
            {prop: value},
            project_entity=convert_value_to_dict(project_entity),
        )

    @property
    def data_type(self):
        """
        :return: The data type of the field.
        :rtype: str
        """
        return self.schema()["data_type"]["value"]

    @data_type.setter
    def data_type(self, value):
        self._update_schema("data_type", value)

    @property
    def description(self):
        """
        :return: The description of the field.
        :rtype: str
        """
        return self.schema()["description"]["value"]

    @property
    def display_name(self):
        """
        :return: The display name of the field.
        :rtype: str
        """
        return self.schema()["name"]["value"]

    @display_name.setter
    def display_name(self, value):
        self._update_schema("name", value)

    @property
    def custom_metadata(self):
        """
        :return: Custom metadata attached to this field.
        :rtype: str
        """
        return self.schema()["custom_metadata"]["value"]

    @custom_metadata.setter
    def custom_metadata(self, value):
        self._update_schema("custom_metadata", value)

    @property
    def properties(self):
        """
        :return: The properties of the field. This strongly depends on the data type of the field.
                 This can for example give you all the possible values of a status field.
        :rtype: dict[str,dict[str,Any]]
        """
        return self.schema()["properties"]

    @properties.setter
    def properties(self, value):
        self._update_schema("properties", value)

    @property
    def valid_types(self):
        """
        :return: The valid SG entity types for entity- and multi-entity-fields.
        :rtype: list[str]
        """
        return self.schema()["properties"]["valid_types"]["value"]

    def batch_update_dict(self, value):
        """
        :param Any value: The value to set.
        :returns: A dict that can be used in a shotgun.batch() call to update this field.
                  Useful when you want to collect field changes and set them in one go.
        :rtype: dict[str,Any]
        """
        value = convert_value_to_dict(value)
        return {
            "request_type": "update",
            "entity_type": self._entity.type,
            "entity_id": self._entity.id,
            "data": {self._name: value},
        }



from .core import convert_fields_to_dicts, convert_fields_to_pysg, new_entity, new_site
from .field import Field


class SGEntity(object):
    """
    An instance of this class represents a single entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, entity_type, entity_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param str entity_type: The ShotGrid type of the entity.
        :param int entity_id: The ID of the ShotGrid entity.
        """
        self._sg = sg  # :type: shotgun_api3.shotgun.Shotgun
        self._type = entity_type
        self._id = entity_id

    def __str__(self):
        return "{} - Type: {} - ID: {} - URL: {}".format(
            self.__class__.__name__, self._type, self._id, self.url
        )

    @property
    def id(self):
        """
        :return: The ID of the ShotGrid entity.
        :rtype: int
        """
        return self._id

    @property
    def type(self):
        """
        :return: The type of the ShotGrid entity.
        :rtype: str
        """
        return self._type

    @property
    def sg(self):
        """
        :return: The Shotgun instance that the entity belongs to.
        :rtype: shotgun_api3.shotgun.Shotgun
        """
        return self._sg

    @property
    def site(self):
        """
        :return: The pyshotgrid site for this entity.
        :rtype: SGSite
        """
        return new_site(self._sg)

    @property
    def url(self):
        """
        :return: The ShotGrid URL for this entity.

            .. note::

               This will only work on entities that have a detail view enabled
               in the system settings.
        :rtype: str
        """
        return "{}/detail/{}/{}".format(self.sg.base_url, self._type, self._id)

    def __getitem__(self, field):
        """
        Enabling dict notation to query fields of the entity from ShotGrid.

        :param str field: The field to query.
        :return: The value of the field. Any entities will be automatically converted to
                 pyshotgrid objects.
        """
        return Field(name=field, entity=self)

    def fields(self, project_entity=None):
        """
        :param dict[str,Any]|SGEntity|None project_entity: A project entity to filter by.
        :return: All fields from this entity. If a project entity is given
                 only fields that are visible to the project are returned.
        :rtype: list[Field]
        """
        if project_entity is not None and not isinstance(project_entity, dict):
            project_entity = project_entity.to_dict()

        sg_entity_fields = self.sg.schema_field_read(
            self._type, project_entity=project_entity
        )
        fields = [
            field
            for field, schema in sg_entity_fields.items()
            if schema["visible"]["value"]
        ]

        return [Field(name=field, entity=self) for field in fields]

    def all_field_values(self, project_entity=None, raw_values=False):
        """
        :param dict[str,Any]|SGEntity project_entity: A project entity to filter by.
        :param bool raw_values: Whether to convert entities to pysg objects or not.
        :return: All fields and values from this entity in a dict. If a project entity is given
                 only fields that are visible to the project are returned.
        :rtype: dict[str,Any]
        """
        if project_entity is not None and not isinstance(project_entity, dict):
            project_entity = project_entity.to_dict()

        sg_entity_fields = self.sg.schema_field_read(
            self._type, project_entity=project_entity
        )
        fields = [
            field
            for field, schema in sg_entity_fields.items()
            if schema["visible"]["value"]
        ]
        all_fields = self.sg.find_one(self._type, [["id", "is", self._id]], fields)

        if raw_values:
            return all_fields
        else:
            return convert_fields_to_pysg(self._sg, all_fields)

    def to_dict(self):
        # noinspection PyUnresolvedReferences
        """
        Creates a dict with just "type" and "id" (and does not call SG).

        Example::

            >>> sg_entity.to_dict()
            {'type': 'CustomEntity01', 'id': 1}

        :returns: The entity as a dict which is ready to consume by the shotgun_api3 methods.

        :rtype: dict[str,Any]
        """
        return {"id": self._id, "type": self._type}

    def batch_update_dict(self, data):
        """
        :param dict[str,Any] data: A dict with the fields and values to set.
        :returns: A dict that can be used in a shotgun.batch() call to update some fields.
                  Useful when you want to collect field changes and set them in one go.
        :rtype: dict[str,Any]
        """
        return {
            "request_type": "update",
            "entity_type": self._type,
            "entity_id": self._id,
            "data": convert_fields_to_dicts(data),
        }

    def set(self, data, multi_entity_update_modes=None):
        """
        Set many fields at once on this entity.

        :param dict[str,Any] data: A dict with the fields and values to set.
        :param dict multi_entity_update_modes: Optional dict indicating what update mode to use
            when updating a multi-entity link field. The keys in the dict are the fields to set
            the mode for, and the values from the dict are one of ``set``, ``add``, or ``remove``.
            Defaults to ``set``.
            ::

                multi_entity_update_modes={"shots": "add", "assets": "remove"}
        :return:
        """
        return self.sg.update(
            self._type,
            self._id,
            data=convert_fields_to_dicts(data),
            multi_entity_update_modes=multi_entity_update_modes,
        )

    def get(self, fields, raw_values=False):
        """
        Set many fields at once on this entity.

        :param list[str] fields: A list of fields to query from this entity.
        :param bool raw_values: Any entities will be converted to pyshotgrid instances.
                                If you set this parameter to True you can turn this behaviour off.
        :return: A dict with the fields and their corresponding values.
        :rtype: dict[str,Any]
        """
        sg_fields = self.sg.find_one(self._type, [["id", "is", self._id]], fields)

        del sg_fields["id"]
        del sg_fields["type"]

        if raw_values:
            return sg_fields
        else:
            return convert_fields_to_pysg(self._sg, sg_fields)

    def delete(self):
        """
        Delete this entity.

        .. Note::

            The python object that represents this entity does not make sense any more after you
            ran this method and will create errors if you keep calling functions on it.

        :return: Whether the entity was successfully deleted.
        :rtype: bool
        """
        return self._sg.delete(self._type, self._id)

    def schema(self):
        """
        :return: The schema for the current entity.
        """
        return self.sg.schema_entity_read()[self._type]

    def field_schemas(self):
        """
        :return: The schemas of all the entity fields.
        """
        return self.sg.schema_field_read(self._type)

    def _publishes(
        self, base_filter=None, pub_types=None, latest=False, additional_sg_filter=None
    ):
        """
        This function is meant as a base for a "publishes" function on a sub class. Publishes
        are stored in different fields for each entity and not every entity has a published file.
        This is why this function is hidden by default.

        :param base_filter: The basic sg filter to get the publishes that are associated with
                            this entity.
        :param str|list[str]|None pub_types: The names of the Publish File Types to return.
        :param bool latest: Whether to get the "latest" publishes or not. This uses the
                            same logic as the tk-multi-loader2 app which is as follows:
                             - group all publishes with the same "name" field together
                             - from these get the publishes with the highest "version_number" field
                             - if there are publishes with the same "name" and "version_number" the
                               newest one wins.
        :param additional_sg_filter:
        :return: All published files from this shot.
        :rtype: list[SGPublishedFile]
        """
        base_filter = base_filter or []
        if pub_types is not None:
            if isinstance(pub_types, list):
                pub_types_filter = {"filter_operator": "any", "filters": []}
                for pub_type in pub_types:
                    pub_types_filter["filters"].append(
                        ["published_file_type.PublishedFileType.code", "is", pub_type]
                    )
            else:
                pub_types_filter = [
                    "published_file_type.PublishedFileType.code",
                    "is",
                    pub_types,
                ]
            base_filter.append(pub_types_filter)

        additional_sg_filter = additional_sg_filter or []
        result_filter = base_filter + additional_sg_filter

        sg_publishes = self.sg.find(
            "PublishedFile", result_filter, ["name", "version_number", "created_at"]
        )
        if latest:
            # group publishes by "name"
            tmp = {}
            for sg_publish in sg_publishes:
                if sg_publish["name"] in tmp:
                    tmp[sg_publish["name"]].append(sg_publish)
                else:
                    tmp[sg_publish["name"]] = [sg_publish]

            # sort them by date and than by version_number which sorts the latest publish to the
            # last position.
            result = []
            for publishes in tmp.values():
                publishes.sort(
                    key=lambda pub: (pub["created_at"], pub["version_number"])
                )
                result.append(publishes[-1])

            # Sort one more time by name.
            result.sort(key=lambda pub: pub["name"])

            sg_publishes = result

        return [new_entity(self._sg, sg_publish) for sg_publish in sg_publishes]



"""
Main module which collects all important functionality.

Use it like::

    >>> import pyshotgrid

"""

from . import sg_default_entities as sde
from .core import new_entity  # noqa: F401
from .core import new_site  # noqa: F401
from .core import (
    register_fallback_pysg_class,
    register_pysg_class,
    register_sg_site_class,
)
from .sg_entity import Field, SGEntity  # noqa: F401
from .sg_site import SGSite  # noqa: F401

#: The pyshotgrid version number as string
VERSION = "0.8.0"

register_sg_site_class(SGSite)
register_fallback_pysg_class(SGEntity)
# Register default pysg plugins
register_pysg_class("Project", sde.SGProject)
register_pysg_class("Shot", sde.SGShot)
register_pysg_class("Asset", sde.SGAsset)
register_pysg_class("Task", sde.SGTask)
register_pysg_class("PublishedFile", sde.SGPublishedFile)
register_pysg_class("Playlist", sde.SGPlaylist)
register_pysg_class("Version", sde.SGVersion)
register_pysg_class("HumanUser", sde.SGHumanUser)



from .core import convert_fields_to_dicts, convert_filters_to_dict, new_entity


class SGSite(object):
    """
    An instance of this class represents a ShotGrid site as a whole.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_site <pyshotgrid.core.new_site>`
        method instead. This gives you more ways to initialize the this class
        and ensures that the plugin system is correctly used.
    """

    def __init__(self, sg):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        """
        self._sg = sg

    @property
    def sg(self):
        """
        :return: The Shotgun instance that the entity belongs to.
        :rtype: shotgun_api3.shotgun.Shotgun
        """
        return self._sg

    def create(self, entity_type, data):
        """
        The same function as
        :py:meth:`Shotgun.create <shotgun_api3:shotgun_api3.shotgun.Shotgun.create>`,
        but it accepts and returns a pyshotgrid object.

        :param str entity_type: The type of the entity to create.
        :param dict[str,Any]|None data: dict of fields and values to set on creation.
                                        The values can contain pysg objects.
        :return: The new created entity.
        :rtype: SGEntity
        """
        # noinspection PyTypeChecker
        return new_entity(
            self._sg,
            self._sg.create(
                entity_type=entity_type,
                data=convert_fields_to_dicts(data),
                return_fields=None,
            ),
        )

    def find(
        self,
        entity_type,
        filters,
        order=None,
        filter_operator=None,
        limit=0,
        retired_only=False,
        page=0,
        include_archived_projects=True,
        additional_filter_presets=None,
    ):
        """
        The same function as
        :py:meth:`Shotgun.find <shotgun_api3:shotgun_api3.shotgun.Shotgun.find>`, but it
        accepts and returns pyshotgrid objects.

        :param entity_type:
        :param filters:
        :param order:
        :param filter_operator:
        :param limit:
        :param retired_only:
        :param page:
        :param include_archived_projects:
        :param additional_filter_presets:
        :return:
        """
        # noinspection PyTypeChecker
        return [
            new_entity(self._sg, sg_entity)
            for sg_entity in self._sg.find(
                entity_type=entity_type,
                filters=convert_filters_to_dict(filters),
                fields=None,
                order=order,
                filter_operator=filter_operator,
                limit=limit,
                retired_only=retired_only,
                page=page,
                include_archived_projects=include_archived_projects,
                additional_filter_presets=additional_filter_presets,
            )
        ]

    def find_one(
        self,
        entity_type,
        filters,
        order=None,
        filter_operator=None,
        limit=0,
        retired_only=False,
        page=0,
        include_archived_projects=True,
        additional_filter_presets=None,
    ):
        """
        The same function as
        :py:meth:`Shotgun.find_one <shotgun_api3:shotgun_api3.shotgun.Shotgun.find_one>` ,
        but it accepts and returns pyshotgrid objects.
        """
        # TODO allow entering the display name for the entity_type
        result = self.find(
            entity_type=entity_type,
            filters=filters,
            order=order,
            filter_operator=filter_operator,
            limit=limit,
            retired_only=retired_only,
            page=page,
            include_archived_projects=include_archived_projects,
            additional_filter_presets=additional_filter_presets,
        )
        if result:
            return result[0]

    def project(self, name_or_id):
        """
        :param int|str name_or_id: The name or id of the project to return.
                                   The name can either match the "tank_name" (recommended)
                                   or the "name" field.
        :return: The found SG project or None.
        :rtype: SGProject|None
        """
        sg_projects = self.projects(names_or_ids=[name_or_id], include_archived=True)
        if sg_projects:
            return sg_projects[0]

    def projects(
        self, names_or_ids=None, include_archived=False, template_projects=False
    ):
        """
        :param list[int|str]|None names_or_ids: List of names or ids of the projects to return. The
                                                names can either match the "tank_name" (recommended)
                                                or the "name" field.
        :param bool include_archived: Whether to include archived projects or not.
        :param bool template_projects: Whether to return template projects or not.
        :return: A list of SG projects.
        :rtype: list[SGProject]
        """
        sg_projects = self._sg.find(
            "Project",
            [["is_template", "is", template_projects]],
            ["tank_name", "name"],
            include_archived_projects=include_archived,
        )

        if names_or_ids is not None:
            if isinstance(names_or_ids[0], int):
                sg_projects = [
                    sg_project
                    for sg_project in sg_projects
                    if sg_project["id"] in names_or_ids
                ]
            else:
                sg_projects = [
                    sg_project
                    for sg_project in sg_projects
                    if (
                        sg_project["tank_name"] in names_or_ids
                        or sg_project["name"] in names_or_ids
                    )
                ]

        return [new_entity(self._sg, sg_project) for sg_project in sg_projects]

    def pipeline_configuration(self, name_or_id=None, project=None):
        """
        :param int|str|None name_or_id: Name or ID of the PipelineConfiguration.
        :param dict|SGEntity|None project: The project that the PipelineConfiguration
                                                 is attached to.
        :return: The PipelineConfiguration.
        :rtype: SGEntity|None
        """
        base_filter = []
        if name_or_id is not None:
            if isinstance(name_or_id, int):
                base_filter = [["id", "is", name_or_id]]
            else:
                base_filter = [["code", "is", name_or_id]]

        if project is not None:
            if isinstance(project, dict):
                base_filter.append(["project", "in", project])
            else:
                base_filter.append(["project", "in", project.to_dict()])

        sg_pipe_config = self._sg.find_one("PipelineConfiguration", base_filter)

        if sg_pipe_config:
            return new_entity(self._sg, sg_pipe_config)

    def people(self, additional_sg_filter=None):
        """
        :param list|None additional_sg_filter:
        :return: All HumanUsers of this ShotGrid site.
        :rtype: list[SGHumanUser]
        """
        # TODO add "only_active" and "name_or_id" parameter
        return [
            new_entity(self._sg, sg_user)
            for sg_user in self._sg.find("HumanUser", additional_sg_filter or [])
        ]



"""
This module collects all default pyshotgrid custom entities.
"""
import fnmatch

from .core import new_entity
from .sg_entity import SGEntity


class SGProject(SGEntity):
    """
    An instance of this class represents a single Project entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, project_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int project_id: The ID of the project entity.
        """
        super(SGProject, self).__init__(sg, entity_type="Project", entity_id=project_id)

    def shots(self, glob_pattern=None):
        """
        :param str|None glob_pattern: A glob to match the shots to return. For example
                                      `TEST_01_*` would return all shots that start with `TEST_01_`.
        :return: All the shots from this project.
        :rtype: list[SGShot]
        """
        sg_shots = self.sg.find("Shot", [["project", "is", self.to_dict()]], ["code"])
        if glob_pattern is not None:
            return [
                new_entity(self._sg, sg_shot)
                for sg_shot in sg_shots
                if fnmatch.fnmatchcase(sg_shot["code"], glob_pattern)
            ]
        else:
            return [new_entity(self._sg, sg_shot) for sg_shot in sg_shots]

    def assets(self, glob_pattern=None):
        """
        :param str|None glob_pattern: A glob to match the assets to return. For example
                                      `TEST_*` would return all assets that start with `TEST_`.
        :return: All the assets from this project.
        :rtype: list[SGAsset]
        """
        sg_assets = self.sg.find("Asset", [["project", "is", self.to_dict()]], ["code"])
        if glob_pattern is not None:
            return [
                new_entity(self._sg, sg_asset)
                for sg_asset in sg_assets
                if fnmatch.fnmatchcase(sg_asset["code"], glob_pattern)
            ]
        else:
            return [new_entity(self._sg, sg_asset) for sg_asset in sg_assets]

    def publishes(self, pub_types=None, latest=False, additional_sg_filter=None):
        """
        :param str|list[str]|None pub_types: The names of the Publish File Types to return.
        :param bool latest: Whether to get the "latest" publishes or not. This uses the
          same logic as the tk-multi-loader2 app which is as follows:

            - group all publishes with the same "name" field together
            - from these get the publishes with the highest "version_number" field
            - if there are publishes with the same "name" and "version_number" the
              newest one wins.
        :param additional_sg_filter:
        :return: All published files from this project.
        :rtype: list[SGPublishedFile]
        """
        return self._publishes(
            base_filter=[["project", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )

    def people(self, additional_sg_filter=None):
        """
        :param list|None additional_sg_filter:
        :return: All HumanUsers assigned to this project.
        :rtype: list[SGHumanUser]
        """
        sg_filter = [["projects", "contains", self.to_dict()]]
        if additional_sg_filter is not None:
            sg_filter += additional_sg_filter

        return [
            new_entity(self._sg, sg_user)
            for sg_user in self._sg.find("HumanUser", sg_filter)
        ]

    def playlists(self):
        """
        :return: All playlists attached to this project.
        :rtype: list[SGPlaylist]
        """
        return [
            new_entity(self._sg, sg_playlist)
            for sg_playlist in self._sg.find(
                "Playlist", [["project", "is", self.to_dict()]]
            )
        ]


class SGShot(SGEntity):
    """
    An instance of this class represents a single Shot entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, shot_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int shot_id: The ID of the shot entity.
        """
        super(SGShot, self).__init__(sg, entity_type="Shot", entity_id=shot_id)

    def publishes(self, pub_types=None, latest=False, additional_sg_filter=None):
        """
        :param str|list[str]|None pub_types: The names of the Publish File Types to return.
        :param bool latest: Whether to get the "latest" publishes or not. This uses the
                            same logic as the tk-multi-loader2 app which is as follows:

                              - group all publishes with the same "name" field together
                              - from these get the publishes with the highest "version_number" field
                              - if there are publishes with the same "name" and "version_number" the
                                newest one wins.
        :param additional_sg_filter:
        :return: All published files from this shot.
        :rtype: list[SGPublishedFile]
        """
        return self._publishes(
            base_filter=[["entity", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )

    def tasks(self, names=None, pipeline_step=None):
        """
        :param list[str]|None names: The names of Tasks to return.
        :param str|dict|SGEntity|None pipeline_step: Name, short name or entity object
                                                     or the Pipeline Step to filter by.
        :returns: A list of Tasks
        :rtype: list[SGTask]
        """
        sg_filter = [["entity", "is", self.to_dict()]]

        if names is not None:
            if len(names) == 1:
                names_filter = ["code", "is", names[0]]
            else:
                names_filter = {"filter_operator": "any", "filters": []}
                for name in names:
                    names_filter["filters"].append(["code", "is", name])
            sg_filter.append(names_filter)

        if pipeline_step is not None:
            if isinstance(pipeline_step, dict):
                sg_filter.append(["step", "is", pipeline_step])
            elif isinstance(pipeline_step, SGEntity):
                sg_filter.append(["step", "is", pipeline_step.to_dict()])
            else:
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["step.Step.code", "is", pipeline_step],
                            ["step.Step.short_name", "is", pipeline_step],
                        ],
                    }
                )

        return [
            new_entity(self._sg, sg_task)
            for sg_task in self._sg.find("Task", sg_filter)
        ]


class SGAsset(SGEntity):
    """
    An instance of this class represents a single Asset entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, asset_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int asset_id: The ID of the Asset entity.
        """
        super(SGAsset, self).__init__(sg, entity_type="Asset", entity_id=asset_id)

    def publishes(self, pub_types=None, latest=False, additional_sg_filter=None):
        """
        :param str|list[str]|None pub_types: The names of the Publish File Types to return.
        :param bool latest: Whether to get the "latest" publishes or not. This uses the
                            same logic as the tk-multi-loader2 app which is as follows:

                              - group all publishes with the same "name" field together
                              - from these get the publishes with the highest "version_number" field
                              - if there are publishes with the same "name" and "version_number" the
                                newest one wins.
        :param additional_sg_filter:
        :return: All published files from this asset.
        :rtype: list[SGPublishedFile]
        """
        return self._publishes(
            base_filter=[["entity", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )

    def tasks(self, names=None, pipeline_step=None):
        """
        :param list[str]|None names: The names of Tasks to return.
        :param str|dict|SGEntity|None pipeline_step: Name, short name or entity object
                                                     or the Pipeline Step to filter by.
        :returns: A list of Tasks
        :rtype: list[SGTask]
        """
        sg_filter = [["entity", "is", self.to_dict()]]

        if names is not None:
            if len(names) == 1:
                names_filter = ["code", "is", names[0]]
            else:
                names_filter = {"filter_operator": "any", "filters": []}
                for name in names:
                    names_filter["filters"].append(["code", "is", name])
            sg_filter.append(names_filter)

        if pipeline_step is not None:
            if isinstance(pipeline_step, dict):
                sg_filter.append(["step", "is", pipeline_step])
            elif isinstance(pipeline_step, SGEntity):
                sg_filter.append(["step", "is", pipeline_step.to_dict()])
            else:
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["step.Step.code", "is", pipeline_step],
                            ["step.Step.short_name", "is", pipeline_step],
                        ],
                    }
                )

        return [
            new_entity(self._sg, sg_task)
            for sg_task in self._sg.find("Task", sg_filter)
        ]


class SGTask(SGEntity):
    """
    An instance of this class represents a single Task entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, task_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int task_id: The ID of the Task entity.
        """
        super(SGTask, self).__init__(sg, entity_type="Task", entity_id=task_id)

    def publishes(self, pub_types=None, latest=False, additional_sg_filter=None):
        """
        :param str|list[str]|None pub_types: The names of the Publish File Types to return.
        :param bool latest: Whether to get the "latest" publishes or not. This uses the
                            same logic as the tk-multi-loader2 app which is as follows:

                             - group all publishes with the same "name" field together
                             - from these get the publishes with the highest "version_number" field
                             - if there are publishes with the same "name" and "version_number" the
                               newest one wins.
        :param additional_sg_filter:
        :return: All published files from this shot.
        :rtype: list[SGPublishedFile]
        """
        return self._publishes(
            base_filter=[["task", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )


class SGPublishedFile(SGEntity):
    """
    An instance of this class represents a single PublishedFile entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, published_file_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int published_file_id: The ID of the PublishedFile entity.
        """
        super(SGPublishedFile, self).__init__(
            sg, entity_type="PublishedFile", entity_id=published_file_id
        )

    # TODO is_latest
    # TODO get_first_publish
    # TODO get_previous_publishes
    # TODO get_next_publishes
    # TODO get_latest_publish
    # TODO get_all_publish_versions


class SGVersion(SGEntity):
    """
    An instance of this class represents a single Version entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, version_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int version_id: The ID of the Version entity.
        """
        super(SGVersion, self).__init__(sg, entity_type="Version", entity_id=version_id)


class SGPlaylist(SGEntity):
    """
    An instance of this class represents a single Playlist entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, playlist_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int playlist_id: The ID of the Playlist entity.
        """
        super(SGPlaylist, self).__init__(
            sg, entity_type="Playlist", entity_id=playlist_id
        )

    @property
    def media_url(self):
        """
        :return: The Media center URL for this playlist.
        :rtype: str
        :raises:
            :RuntimeError: When this playlist is not attached to a project.
        """
        sg_project = self["project"].get()
        if sg_project is None:
            raise RuntimeError(
                'Cannot get media URL for playlist "{}"'
                ", because it is not attached to a project.".format(self.id)
            )
        # Example URL:
        # https://example.shotgunstudio.com/page/media_center?type=Playlist&id=123&project_id=456
        return "{}/page/media_center?type={}&id={}&project_id={}".format(
            self.sg.base_url, self._type, self._id, sg_project["id"].get()
        )


class SGHumanUser(SGEntity):
    """
    An instance of this class represents a single HumanUser entity in ShotGrid.

    .. Note::

        Try to avoid creating instances of this class in production code and
        use the :py:meth:`pyshotgrid.new_entity <pyshotgrid.core.new_entity>`
        method instead. This will make sure that you always get the correct
        entity class to work with.
    """

    def __init__(self, sg, human_user_id):
        """
        :param shotgun_api3.shotgun.Shotgun sg:
            A fully initialized instance of shotgun_api3.Shotgun.
        :param int human_user_id: The ID of the PublishedFile entity.
        """
        super(SGHumanUser, self).__init__(
            sg, entity_type="HumanUser", entity_id=human_user_id
        )

    def tasks(self, names=None, project=None, pipeline_step=None):
        """
        :param list[str]|None names: The names of Tasks to return.
        :param str|dict|SGEntity|None project: Name, tank_name or entity object
                                               or the Project to filter by.
        :param str|dict|SGEntity|None pipeline_step: Name, short name or entity object
                                                     or the Pipeline Step to filter by.
        :returns: A list of Tasks
        :rtype: list[SGTask]
        """
        sg_filter = [
            {
                "filter_operator": "any",
                "filters": [
                    ["task_assignees", "contains", self.to_dict()],
                    ["task_assignees.Group.users", "contains", self.to_dict()],
                ],
            }
        ]

        if names is not None:
            if len(names) == 1:
                names_filter = ["code", "is", names[0]]
            else:
                names_filter = {"filter_operator": "any", "filters": []}
                for name in names:
                    names_filter["filters"].append(["code", "is", name])
            sg_filter.append(names_filter)

        if project is not None:
            if isinstance(project, dict):
                sg_filter.append(["project", "is", project])
            elif isinstance(project, SGEntity):
                sg_filter.append(["project", "is", project.to_dict()])
            else:
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["project.Project.code", "is", project],
                            ["project.Project.tank_name", "is", project],
                        ],
                    }
                )

        if pipeline_step is not None:
            if isinstance(pipeline_step, dict):
                sg_filter.append(["step", "is", pipeline_step])
            elif isinstance(pipeline_step, SGEntity):
                sg_filter.append(["step", "is", pipeline_step.to_dict()])
            else:
                sg_filter.append(
                    {
                        "filter_operator": "any",
                        "filters": [
                            ["step.Step.code", "is", pipeline_step],
                            ["step.Step.short_name", "is", pipeline_step],
                        ],
                    }
                )

        return [
            new_entity(self._sg, sg_task)
            for sg_task in self._sg.find("Task", sg_filter)
        ]

    def publishes(self, pub_types=None, latest=False, additional_sg_filter=None):
        """
        :param str|list[str]|None pub_types: The names of the Publish File Types to return.
        :param bool latest: Whether to get the "latest" publishes or not. This uses the
                            same logic as the tk-multi-loader2 app which is as follows:

                             - group all publishes with the same "name" field together
                             - from these get the publishes with the highest "version_number" field
                             - if there are publishes with the same "name" and "version_number" the
                               newest one wins.
        :param additional_sg_filter:
        :return: All published files from this shot.
        :rtype: list[SGPublishedFile]
        """
        return self._publishes(
            base_filter=[["created_by", "is", self.to_dict()]],
            pub_types=pub_types,
            latest=latest,
            additional_sg_filter=additional_sg_filter,
        )
